# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

import json
import re
from time import time
from typing import List, Optional, Pattern
import usagelogger
from usagelogger.http_rule import HttpRule


class MetaHttpLogger(type):
    """
    Implement a proper getter/setter for default_rules class variable
    using a metaclass since Python properties only work for instance
    variables by default.
    """
    _default_rules: str = ''

    @property
    def default_rules(cls) -> str:
        return cls._default_rules

    @default_rules.setter
    def default_rules(cls, rules: str) -> None:
        cls._default_rules: str = re.sub(r'^\s*include default\s*$', '',
                                         rules, flags=re.MULTILINE)


class HttpLogger(usagelogger.BaseLogger, metaclass=MetaHttpLogger):
    """Usage logger for HTTP/HTTPS protocol."""

    # Agent string identifying this logger.
    AGENT: str = 'http_logger.py'

    _default_rules: str = usagelogger.HttpRules.strict_rules()

    # To avoid mistakes ensure self.default_rules == HttpLogger.default_rules
    # because... Python
    @property
    def default_rules(self) -> str:
        return str(HttpLogger.default_rules)

    @default_rules.setter
    def default_rules(self, value: str) -> None:
        HttpLogger.default_rules = value

    @staticmethod
    def is_string_content_type(content: str) -> bool:
        string_types: str = (
            r'^(text/(html|plain|xml))|'
            r'(application/(json|soap|xml|x-www-form-urlencoded))')
        pattern: Pattern = re.compile(string_types, flags=re.IGNORECASE)
        return content is not None and pattern.match(content) is not None

    def __init__(self, _agent_unused=None,
                 enabled: Optional[bool] = True,
                 queue: Optional[List[str]] = None,
                 url: Optional[str] = usagelogger.UsageLoggers.url_by_default(),
                 skip_compression: Optional[bool] = False,
                 skip_submission: Optional[bool] = False,
                 rules: Optional[str] = None) -> None:
        """
        Initialize enabled/disabled logger
        using specified url and default rules.
        """

        super().__init__(self.AGENT, enabled=enabled, queue=queue, url=url,
                         skip_compression=skip_compression,
                         skip_submission=skip_submission)

        if rules is not None:
            self.rules: str = re.sub(r'^\s*include default\s*$',
                                     str(self.default_rules),
                                     rules, flags=re.MULTILINE)
            if len(self.rules.strip()) == 0:
                self.rules: str = self.default_rules
        else:
            self.rules: str = self.default_rules

        # parse and break rules out by verb
        prs: List[HttpRule] = usagelogger.HttpRules.parse(self.rules)
        self.rules_allow_http_url: bool = len([
            r for r in prs if 'allow_http_url' == r.verb]) > 0
        self.rules_copy_session_field: List[HttpRule] = [
            r for r in prs if 'rs_copy_session_field' == r.verb]
        self.rules_remove: List[HttpRule] = [
            r for r in prs if 'rules_remove' == r.verb]
        self.rules_remove_if: List[HttpRule] = [
            r for r in prs if 'rules_remove_if' == r.verb]
        self.rules_remove_if_found: List[HttpRule] = [
            r for r in prs if 'rules_remove_if_found' == r.verb]
        self.rules_remove_unless: List[HttpRule] = [
            r for r in prs if 'rules_remove_unless' == r.verb]
        self.rules_remove_unless_found: List[HttpRule] = [
            r for r in prs if 'rules_remove_unless_found' == r.verb]
        self.rules_replace: List[HttpRule] = [
            r for r in prs if 'rules_replace' == r.verb]
        self.rules_sample: List[HttpRule] = [
            r for r in prs if 'rules_sample' == r.verb]
        self.rules_stop: List[HttpRule] = [
            r for r in prs if 'rules_stop' == r.verb]
        self.rules_stop_if: List[HttpRule] = [
            r for r in prs if 'rules_stop_if' == r.verb]
        self.rules_stop_if_found: List[HttpRule] = [
            r for r in prs if 'rules_stop_if_found' == r.verb]
        self.rules_stop_unless: List[HttpRule] = [
            r for r in prs if 'rules_stop_unless' == r.verb]
        self.rules_stop_unless_found: List[HttpRule] = [
            r for r in prs if 'rules_stop_unless_found' == r.verb]
        self.skip_compression: bool = len([
            r for r in prs if r.verb == 'skip_compression']) > 0
        self.skip_submission: bool = len([
            r for r in prs if r.verb == 'skip_submission']) > 0

        # finish validating rules
        if len(self.rules_sample) > 1:
            raise SyntaxError('Multiple sample rules')
        if self._enabled and url is not None \
                and url.startswith('http:') \
                and not self.rules_allow_http_url:
            self._enableable = False
            self._enabled = False

    def log(self, request: usagelogger.HttpRequestImpl,
            response: usagelogger.HttpResponseImpl,
            response_body: Optional[str] = None,
            request_body: Optional[str] = None) -> bool:
        """Logs HTTP request and response to intended destination."""

        if self.enabled:
            return self.submit(
                self.format(request, response, response_body, request_body))
        return True

    def format(self, request: usagelogger.HttpRequestImpl,
               response: usagelogger.HttpResponseImpl,
               response_body: Optional[str] = None,
               request_body: Optional[str] = None,
               now: Optional[int] = None) -> str:
        """Formats HTTP request and response as JSON message."""

        details: List[List[str]] = usagelogger.HttpMessage.build(
            request, response, response_body, request_body)

        # TODO: copy data from session if configured
        # TODO: quit early based on stop rules if configured
        # TODO: do sampling if configured
        # TODO: winnow sensitive details based on remove rules if configured
        # TODO: mask sensitive details based on replace rules if configured
        # TODO: remove any details with empty values

        details.append(['now', str(now)if now is not None else str(
            round(time() * 1000))])
        details.append(['agent', self.AGENT])
        details.append(['version', self.version])
        return json.dumps(details, separators=(',', ':'))



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

from typing import Dict, Optional, Union


class HttpResponseImpl(object):

    def __init__(self, status: Optional[int] = None,
                 headers: Optional[Dict[str, str]] = None,
                 body: Optional[Union[bytes, str]] = None) -> None:

        self.reason: str = ''

        self.status = status
        self.headers = {} if headers is None else headers
        self.body = body

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value: Optional[Union[bytes, str]]):
        if not value:
            self._body = None
        elif isinstance(value, str):
            self._body = bytes(value, 'utf8')
        elif isinstance(value, bytes):
            self._body = value
        else:
            self._body = None

    def get_headers(self):
        return self.headers.items()



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

from typing import Dict, List, ItemsView, Optional, Set, Tuple, Union
from urllib.parse import parse_qsl, urlsplit, SplitResult


class HttpRequestImpl(object):
    """
    HTTP Request Object

    :param method (Optional): HTTP request method; see valid methods below
    :param request_url (Optional): Full request URL including query string
    :param headers (Optional): HTTP request headers
    :param body (Optional): HTTP request body
    :return: :class: HttpRequestImpl object
    """

    VALID_REQUEST_METHODS: Set[str] = {'GET', 'HEAD', 'POST', 'PUT', 'DELETE',
                                       'PATCH', 'TRACE', 'OPTIONS', 'CONNECT'}
    VALID_SCHEMES: Dict[str, int] = {'https': 443, 'http': 80}

    def __init__(self, method: Optional[str] = None,
                 request_url: Optional[str] = None,
                 headers: Optional[Dict[str, str]] = None,
                 body: Optional[Union[bytes, str]] = None) -> None:

        self._scheme: Optional[str] = None
        self._port: Optional[int] = None

        self.host: str = ''
        self.path: str = ''
        self.query_string: str = ''

        self.method = method
        self.request_url = request_url
        self.headers = {} if headers is None else headers
        self.body = body

    # self.method getter/setter
    @property
    def method(self) -> Optional[str]:
        return self._method

    @method.setter
    def method(self, value: Optional[str]) -> None:
        if value is None or value in self.VALID_REQUEST_METHODS:
            self._method = value
        else:
            raise TypeError('incorrect HTTP method')

    # self.request_url getter/setter
    @property
    def request_url(self) -> Optional[str]:
        """Rebuilds and returns the full URL using base_url, full_path"""
        url_builder = ''
        if self.base_url:
            url_builder = url_builder + self.base_url
            if self.full_path:
                url_builder = url_builder + self.full_path
        return url_builder

    @request_url.setter
    def request_url(self, value: str) -> None:
        """Splits given URL and updates base_url, path, query_string"""
        if value:
            u: SplitResult = urlsplit(value)
            self.base_url = '{0}://{1}'.format(u.scheme, u.netloc)
            self.path = u.path
            self.query_string = u.query

    # self.body getter/setter
    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value: Optional[Union[bytes, str]]):
        if not value:
            self._body = None
        elif isinstance(value, str):
            self._body = bytes(value, 'utf8')
        elif isinstance(value, bytes):
            self._body = value
        else:
            self._body = None

    # self.base_url getter/setter
    @property
    def base_url(self) -> str:
        """Rebuilds and returns the base URL using scheme, host, port"""
        url_builder = ''
        if self._scheme and self.host:
            url_builder = "{0}://{1}".format(self._scheme, self.host)
            if self._port and self._port not in {80, 443}:
                url_builder = url_builder + ":{0}".format(str(self._port))
        return url_builder

    @base_url.setter
    def base_url(self, value: str) -> None:
        """Splits URL and updates scheme, host, port"""
        u: SplitResult = urlsplit(value)
        if u.netloc:
            self.scheme = u.scheme
            self.host = u.hostname
            self.port = self._port_matcher(u.port, u.scheme)
        else:
            raise TypeError('invalid base URL')

    # self.scheme getter/setter
    @property
    def scheme(self) -> Optional[str]:
        return self._scheme

    @scheme.setter
    def scheme(self, value: str) -> None:
        """Verifies scheme is valid and updates scheme, port (when needed)"""
        if value in self.VALID_SCHEMES:
            self._scheme = value
            # Update port if we are using a common port
            if self._port in self.VALID_SCHEMES.values():
                self._port = self._port_matcher(None, self._scheme)
        else:
            raise TypeError('incorrect URL scheme')

    # self.port getter/setter
    @property
    def port(self) -> Optional[int]:
        return self._port

    @port.setter
    def port(self, value: int) -> None:
        """Updates port, scheme (when needed)"""
        if isinstance(value, int):
            self._port = value
            # Update scheme in case we switched to a common port
            if self._port in self.VALID_SCHEMES.values():
                self._scheme = self._scheme_matcher(self._port)
        else:
            raise TypeError('invalid HTTP port')

    # self.full_path getter
    @property
    def full_path(self) -> Optional[str]:
        """Rebuilds and returns the full path using path, query_string"""
        path_builder = ''
        if self.path:
            path_builder = path_builder + self.path
        # Unlike URL methods, process query string even if path is blank
        if self.query_string:
            path_builder = "{0}?{1}".format(path_builder, self.query_string)
        return path_builder

    # formatters
    def get_headers(self) -> ItemsView[str, str]:
        return self.headers.items()

    def get_query_params(self) -> List[Tuple[str, str]]:
        return parse_qsl(self.query_string, keep_blank_values=True)

    # Private
    def _scheme_matcher(self, port: Union[int, str]):
        return next((
            s for s, p in self.VALID_SCHEMES.items() if p == int(port)), None)

    def _port_matcher(self, port: Optional[Union[int, str]],
                      scheme: str) -> Optional[int]:
        if port is None:
            return next((
                p for s, p in self.VALID_SCHEMES.items() if s == scheme), None)
        else:
            return int(port)



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

import json
import http.client
from typing import Dict, List, Optional
from urllib.parse import urlsplit
import usagelogger
from usagelogger.usage_loggers import UsageLoggers


class BaseLogger(object):
    """Basic usage logger to embed or extend."""

    def __init__(self, agent: str,
                 enabled: Optional[bool] = True,
                 queue: Optional[List[str]] = None,
                 url: Optional[str] = UsageLoggers.url_by_default(),
                 skip_compression: Optional[bool] = False,
                 skip_submission: Optional[bool] = False) -> None:

        self.agent = agent
        self.skip_compression = skip_compression
        self.skip_submission = skip_submission
        self.version = self.version_lookup()

        # set options in priority order
        self._enabled = enabled
        self._queue = queue if isinstance(queue, list) else None
        if self._queue is not None:
            self._url = None
        elif url is not None and isinstance(url, str):
            try:
                if urlsplit(url).scheme in {'http', 'https'}:
                    self._url_scheme: str = urlsplit(url).scheme
                    self._url = url
                else:
                    raise TypeError('incorrect URL scheme')
            except TypeError:
                self._enabled = False
                self._url = None
        else:
            self._enabled = False
            self._url = None
        self._enableable = self.queue is not None or self.url is not None

    def disable(self):
        self._enabled = False
        return self

    def enable(self):
        if self.enableable:
            self._enabled = True
        return self

    @property
    def enableable(self) -> bool:
        return self._enableable

    @property
    def enabled(self) -> bool:
        return self._enabled and UsageLoggers.is_enabled()

    @property
    def queue(self) -> List[str]:
        return self._queue

    def submit(self, msg: Optional[str]) -> bool:
        """Submits JSON message to intended destination."""

        if msg is None or self.skip_submission is True or self.enabled is False:
            return True
        elif self._queue is not None:
            self._queue.append(msg)
            return True
        else:
            try:
                # TODO: implement compression
                url_parser = urlsplit(self.url)
                hostname = url_parser.hostname  # todo cache this!
                url_path = url_parser.path + url_parser.query  # cache todo this!

                body: str = json.dumps(msg).encode('utf8')
                headers: Dict[str, str] = {'Content-Type': 'application/json'}

                if self._url_scheme == "http":
                    conn = http.client.HTTPConnection(hostname)
                else:
                    conn = http.client.HTTPSConnection(hostname)

                conn.request("POST", url_path, body, headers)
                response = conn.getresponse()
                conn.close()  # todo keep alive? (or should be closed in except handler?)

                return response.status == 204

            # http errors
            except (http.client.HTTPException, IOError, OSError):
                return False
            # JSON errors
            except (OverflowError, TypeError, ValueError):
                return False

    @property
    def url(self) -> str:
        return self._url

    @staticmethod
    def version_lookup() -> str:
        return usagelogger.__version__



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

import re
from typing import List, Pattern
from usagelogger.http_rule import HttpRule


class HttpRules(object):

    @classmethod
    def debug_rules(cls) -> str:
        return (r"allow_http_url" "\n"
                r"copy_session_field /.*/" "\n")

    @classmethod
    def standard_rules(cls) -> str:
        return ("/request_header:cookie|response_header:set-cookie/ remove\n"
                r'/(request|response)_body|request_param/ replace '
                r'/[a-zA-Z0-9.!#$%&’*+\/=?^_`{|}~-]+@[a-zA-Z0-9-]+'
                r'(?:\.[a-zA-Z0-9-]+)/, /x@y.com/' "\n"
                r'/request_body|request_param|response_body/ replace '
                r'/[0-9\.\-\/]{9,}/, /xyxy/' "\n")

    @classmethod
    def strict_rules(cls) -> str:
        return (r'/request_url/ replace /([^\?;]+).*/, !\\1!' "\n"
                r'/request_body|response_body|request_param:.*|'
                r'request_header:(?!user-agent).*|response_header:'
                r'(?!(content-length)|(content-type)).*/ remove' "\n")

    @classmethod
    def parse(cls, rules: str) -> List[HttpRule]:
        """Parses rules from multi-line string."""
        result: List = []
        if rules is not None:
            rules = re.sub(r'(?m)^\s*include debug\s*$', HttpRules.debug_rules(), rules)
            rules = re.sub(r'(?m)^\s*include standard\s*$', HttpRules.standard_rules(), rules)
            rules = re.sub(r'(?m)^\s*include strict\s*$', HttpRules.strict_rules(), rules)
            for rule in rules.split("\n"):
                parsed: HttpRule = HttpRules.parse_rule(rule)
                if parsed is not None: result.append(parsed)
        return result

    @classmethod
    def parse_rule(cls, rule: str) -> HttpRule:
        """Parses rule from single line."""
        if rule is None or HttpRules._REGEX_BLANK_OR_COMMENT.match(rule): return None
        m = HttpRules._REGEX_ALLOW_HTTP_URL.match(rule)
        if m: return HttpRule('allow_http_url')
        m = HttpRules._REGEX_COPY_SESSION_FIELD.match(rule)
        if m: return HttpRule('copy_session_field', None, HttpRules._parse_regex(rule, m.group(1)))
        m = HttpRules._REGEX_REMOVE.match(rule)
        if m: return HttpRule('remove', HttpRules._parse_regex(rule, m.group(1)))
        m = HttpRules._REGEX_REMOVE_IF.match(rule)
        if m: return HttpRule('remove_if', HttpRules._parse_regex(rule, m.group(1)), HttpRules._parse_regex(rule, m.group(2)))
        m = HttpRules._REGEX_REMOVE_IF_FOUND.match(rule)
        if m: return HttpRule('remove_if_found', HttpRules._parse_regex(rule, m.group(1)),
                              HttpRules._parse_regex_find(rule, m.group(2)))
        m = HttpRules._REGEX_REMOVE_UNLESS.match(rule)
        if m: return HttpRule('remove_unless', HttpRules._parse_regex(rule, m.group(1)), HttpRules._parse_regex(rule, m.group(2)))
        m = HttpRules._REGEX_REMOVE_UNLESS_FOUND.match(rule)
        if m: return HttpRule('remove_unless_found', HttpRules._parse_regex(rule, m.group(1)),
                              HttpRules._parse_regex_find(rule, m.group(2)))
        m = HttpRules._REGEX_REPLACE.match(rule)
        if m: return HttpRule('replace', HttpRules._parse_regex(rule, m.group(1)), HttpRules._parse_regex_find(rule, m.group(2)),
                              HttpRules._parse_string(rule, m.group(3)))
        m = HttpRules._REGEX_SAMPLE.match(rule)
        if m:
            m1 = int(m.group(1))
            if m1 < 1 or m1 > 99: raise SyntaxError('Invalid sample percent: ' + m.group(1))
            return HttpRule('sample', None, m1)
        m = HttpRules._REGEX_SKIP_COMPRESSION.match(rule)
        if m: return HttpRule('skip_compression')
        m = HttpRules._REGEX_SKIP_SUBMISSION.match(rule)
        if m: return HttpRule('skip_submission')
        m = HttpRules._REGEX_STOP.match(rule)
        if m: return HttpRule('stop', HttpRules._parse_regex(rule, m.group(1)))
        m = HttpRules._REGEX_STOP_IF.match(rule)
        if m: return HttpRule('stop_if', HttpRules._parse_regex(rule, m.group(1)), HttpRules._parse_regex(rule, m.group(2)))
        m = HttpRules._REGEX_STOP_IF_FOUND.match(rule)
        if m: return HttpRule('stop_if_found', HttpRules._parse_regex(rule, m.group(1)),
                              HttpRules._parse_regex_find(rule, m.group(2)))
        m = HttpRules._REGEX_STOP_UNLESS.match(rule)
        if m: return HttpRule('stop_unless', HttpRules._parse_regex(rule, m.group(1)), HttpRules._parse_regex(rule, m.group(2)))
        m = HttpRules._REGEX_STOP_UNLESS_FOUND.match(rule)
        if m: return HttpRule('stop_unless_found', HttpRules._parse_regex(rule, m.group(1)),
                              HttpRules._parse_regex_find(rule, m.group(2)))
        raise SyntaxError('Invalid rule: ' + rule)

    @classmethod
    def _parse_regex(cls, rule: str, regex: str) -> Pattern:
        """Parses regex for matching."""
        s: str = HttpRules._parse_string(rule, regex)
        if not s.startswith('^'): s = '^' + s
        if not s.endswith('$'): s = s + '$'
        try:
            return re.compile(s)
        except:
            raise SyntaxError('Invalid regex ({regex}) in rule: {r}'.format(r=rule, regex=regex))

    @classmethod
    def _parse_regex_find(cls, rule: str, regex: str) -> Pattern:
        """Parses regex for finding."""
        try:
            return re.compile(HttpRules._parse_string(rule, regex)) # todo need flags here?
        except:
            raise SyntaxError('Invalid regex ({regex}) in rule: {r}'.format(r=rule, regex=regex))

    @classmethod
    def _parse_string(cls, rule: str, string: str) -> str:
        """Parses delimited string expression."""
        for sep in ['~', '!', '%', '|', '/']:
            p: Pattern = re.compile(r'^[{s}](.*)[{s}]$'.format(s=sep))
            m = p.match(string)
            if m:
                m1: str = m.group(1)
                m1p: Pattern = re.compile(r'^[{s}].*|.*[^\\\\][{s}].*'.format(s=sep))
                if m1p.match(m1): raise SyntaxError('Unescaped separator ({s}) in rule: {r}'.format(r=rule, s=sep))
                return sep.join(m1.split('\\' + sep))
        raise SyntaxError('Invalid expression ({s}) in rule: {r}'.format(r=rule, s=string))

    _REGEX_ALLOW_HTTP_URL: Pattern = re.compile(
        r'^\s*allow_http_url\s*(#.*)?$')
    _REGEX_BLANK_OR_COMMENT: Pattern = re.compile(
        r'^\s*([#].*)*$')
    _REGEX_COPY_SESSION_FIELD: Pattern = re.compile(
        r'^\s*copy_session_field\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_REMOVE: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*remove\s*(#.*)?$')
    _REGEX_REMOVE_IF: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*remove_if\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_REMOVE_IF_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_if_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_REMOVE_UNLESS: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_unless\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_REMOVE_UNLESS_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_unless_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_REPLACE: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'replace[\s]+([~!%|/].+[~!%|/]),[\s]+([~!%|/].*[~!%|/])\s*(#.*)?$')
    _REGEX_SAMPLE: Pattern = re.compile(
        r'^\s*sample\s+(\d+)\s*(#.*)?$')
    _REGEX_SKIP_COMPRESSION: Pattern = re.compile(
        r'^\s*skip_compression\s*(#.*)?$')
    _REGEX_SKIP_SUBMISSION: Pattern = re.compile(
        r'^\s*skip_submission\s*(#.*)?$')
    _REGEX_STOP: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop\s*(#.*)?$')
    _REGEX_STOP_IF: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop_if\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_STOP_IF_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'stop_if_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_STOP_UNLESS: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop_unless\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    _REGEX_STOP_UNLESS_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'stop_unless_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

import os


class UsageLoggers(object):

    DISABLED: bool = os.getenv('USAGE_LOGGERS_DISABLE') == 'True'
    _disabled: bool = DISABLED

    @classmethod
    def disable(cls) -> None:
        cls._disabled = True

    @classmethod
    def enable(cls) -> None:
        if cls.DISABLED is False:
            cls._disabled = False

    @classmethod
    def is_enabled(cls) -> bool:
        return not cls._disabled

    @staticmethod
    def url_by_default() -> str:
        return os.getenv('USAGE_LOGGERS_URL')



# coding: utf-8
# © 2016-2019 Resurface Labs Inc.

from typing import List, Optional
from usagelogger import HttpRequestImpl, HttpResponseImpl


class HttpMessage(object):

    @classmethod
    def build(cls, request: HttpRequestImpl, response: HttpResponseImpl,
              response_body: Optional[str] = None, request_body: Optional[str] = None):
        message: List[List[str]] = []
        if request.method: message.append(['request_method', request.method])
        if request.request_url: message.append(['request_url', request.request_url])
        if response.status: message.append(['response_code', str(response.status)])
        cls._append_request_headers(message, request)
        cls._append_request_params(message, request)
        cls._append_response_headers(message, response)
        final_request_body = request_body if (request_body is not None) else request.body
        if final_request_body: message.append(['request_body', final_request_body])
        final_response_body = response_body if (response_body is not None) else response.body
        if final_response_body: message.append(['response_body', final_response_body])
        return message

    @classmethod
    def _append_request_headers(cls, message: List[List[str]],
                                request: HttpRequestImpl) -> None:
        for k, v in request.get_headers():
            message.append(["request_header:{0}".format(k).lower(), v])

    @classmethod
    def _append_request_params(cls, message: List[List[str]],
                               request: HttpRequestImpl) -> None:
        # todo read from params dict
        return

    @classmethod
    def _append_response_headers(cls, message: List[List[str]],
                                 response: HttpResponseImpl) -> None:
        for k, v in response.get_headers():
            message.append(["response_header:{0}".format(k).lower(), v])
